#include "rpi_lib/rpi.h"
#include <stdio.h>
#include <stdint.h>

#include "Ab_192.h"

double c[SIZE];
double x[SIZE];

extern void enable_fpu(void);

void spi_print(char *s);
void Serial_send(uint8_t *buf, int len);
void Serial_receive(uint8_t *buf, int len);

char str[100];

// #define DEBUG


void myLUsolver_node0(void){
	// 外積方式ガウス法
	
	int i,j,k;
	int n = SIZE;

	// LU分解
	for (k=0; k<n/2; k++) {
		// kで回す行方向ループ

#ifdef DEBUG
		sprintf(str,"%d, ",k);
		spi_print(str);
#endif

		// 行列Lの処理
		double div_tmp = 1.0 / A[k][k];

		for (i=k+1; i<n; i++) {
			// iで回す行方向ループ

			// 開始メッセージ送信
			uart0_putc('s');

			// l_ik = a_ik / a_kk
			A[i][k] = A[i][k] * div_tmp;
			// 枢軸ベクトルを送信
			Serial_send((uint8_t *)&A[i][k],sizeof(double)*1);

			// 終了メッセージ待ち
			while(uart0_getc() != 'e');
		}

		// 行列Uの処理
		for (j=k+1; j<n; j++) {
			// jで回す行方向のループ

			// Tips: ここで値を変数に入れたほうが
			// ループ内で参照するよりメモリアクセス回数が減ってお得
			double a_jk = A[j][k];

			for (i=k+1; i<n/2; i++) {
				// iで回す列方向のループ
				A[j][i] = A[j][i] - A[k][i] * a_jk; 
			}
		}
	}
	// LU分解終了

#ifdef DEBUG
	sprintf(str,"\n");
	spi_print(str);
#endif

	// 終了待ち
	while(uart0_getc() != 'e');

	// k/2列目以降のデータを受信
	for(i=0;i<n;i++){

#ifdef DEBUG
		sprintf(str,"%d, ",i);
		spi_print(str);
#endif

		for(j=n/2;j<n;j++){
			// 開始メッセージ待ち
			while(uart0_getc() != 's');
			Serial_receive((uint8_t *)&A[i][j],sizeof(double)*1);
			// 終了メッセージ送信
			uart0_putc('e');
		}
	}

#ifdef DEBUG
	sprintf(str,"\n");
	spi_print(str);
#endif


	// 前進代入
	// Lc = b
	for(k=0; k<n; k++){
		// 行方向のループ
		c[k] = b[k];
		for(i=0; i<k; i++){
			// 列方向のループ
			c[k] -= A[k][i] * c[i];
		}
	}


	// 後進代入
	// Ux = c
	// k = n - 1
	x[n-1] = c[n-1] / A[n-1][n-1];
	// k = (n - 2) ~ (0)
	for(k = n - 2; k >= 0; k--){
		x[k] = c[k];
		for(i=k+1; i<n; i++){
			// 列方向のループ
			x[k] -= A[k][i] * x[i];
		}
		x[k] = x[k] / A[k][k];
	}
}


void spi_print(char *s){
	while(*s != '\0'){
		SPI0_transfer(*s);
		s++;
	}
}

void Serial_send(uint8_t *buf, int len){
	int i;
	for(i=0;i<len;i++){
		// 送信FIFOが空くのを待つ
		while(*UART0_FR & (0x01 << 5));
		// データ送信
		int c = buf[i];
		*UART0_DR = c;
	}
}

void Serial_receive(uint8_t *buf, int len){
	int i;
	for(i=0;i<len;i++){
		// データが来るまで待つ
		// RXFE: RX FIFOが空の時に1が立つ
		while(*UART0_FR & (0x01 << 4));
		// データ受信
		int c = *UART0_DR;
		buf[i] = c & 0xff;
	}
}


int main(void){
	int i;

	rpi_init();
	enable_fpu();

	// デバッグログ出力設定
	SPI0_begin(SPI_SS_CS0);
	SPI0_setDataMode(SPI0_CS_MODE_00);

	// UARTボーレート設定
	// baudrate = F_UART / (16*BAUDDIV)
	// baudrate_max = F_UART / (16*0.0625) = F_URAT
	// BAUDDIV = IBRD + ((FBRD - 0.5) / 64)
	// FBRD = int(0.0625 * 64 + 0.5) = 4
	*UART0_IBRD = 0;
	*UART0_FBRD = 0;


	sprintf(str,"start\n");
	spi_print(str);

	delay(1000);


	uint64_t start_time = get_systime();

	myLUsolver_node0();

	uint64_t end_time = get_systime();

	// 実行時間出力
	sprintf(str,"start time: %lld\n",start_time);
	spi_print(str);
	sprintf(str,"end time: %lld\n",end_time);
	spi_print(str);
	sprintf(str,"exec time: %lld\n",end_time - start_time);
	spi_print(str);


	// 結果出力
	sprintf(str,"x: ");
	spi_print(str);
	for(i=0;i<SIZE;i++){
		sprintf(str,"% 2.2f ",x[i]);
		spi_print(str);
	}
	sprintf(str,"\n");
	spi_print(str);


	return 0;
}
